@@ -1,13 +1,14 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2007 Dalius Dobravolskas <dalius@sandbox.lt>
+# Copyright (C) 2007-2013 Dalius Dobravolskas and Geoffrey T. Dairiki
 # All rights reserved.
 #
 # This software is licensed using the same licence as Trac:
 # http://trac.edgewall.org/wiki/TracLicense.
 #
-# Author: Dalius Dobravolskas <dalius@sandbox.lt>
-#
+# Original Author: Dalius Dobravolskas <dalius@sandbox.lt>
+# Current Maintainer: Jeff Dairiki <dairiki@dairiki.org>
+
 # Most probably you will want to add following lines to your configuration file:
 #
 #   [components]
@@ -22,10 +23,11 @@ import time
 import itertools
 
 from trac.core import *
-from trac.config import Option, BoolOption, IntOption
+from trac.config import Option, BoolOption, ListOption
 from trac.web.chrome import INavigationContributor, ITemplateProvider, add_stylesheet, add_script
 from trac.env import IEnvironmentSetupParticipant
 from trac.web.main import IRequestHandler, IAuthenticator
+from trac.perm import IPermissionGroupProvider, PermissionSystem
 from trac.web.session import DetachedSession
 try:
     from acct_mgr.web_ui import LoginModule
@@ -40,8 +42,11 @@ from openid.store.memstore import MemoryStore
 
 from openid.consumer import consumer
 from openid.extensions import sreg, pape, ax
-
 from openid import oidutil
+try:
+    from openid_teams.teams import TeamsRequest, TeamsResponse
+except ImportError:
+    TeamsRequest = TeamsResponse = None
 
 import socket
 import struct
@@ -73,7 +78,8 @@ class AuthOpenIdPlugin(Component):
         lastname='http://axschema.org/namePerson/last',
         )
 
-    implements(INavigationContributor, IRequestHandler, ITemplateProvider, IAuthenticator, IEnvironmentSetupParticipant)
+    implements(INavigationContributor, IRequestHandler, ITemplateProvider, IAuthenticator, IEnvironmentSetupParticipant,
+        IPermissionGroupProvider)
 
     # Do not declare options in the [trac] section.  We should not
     # be creating new declared options there, and we should not be
@@ -132,6 +138,18 @@ class AuthOpenIdPlugin(Component):
             is backwards-incompatible if you already have user sessions
             which you would like to preserve. """)
 
+    use_nickname_as_authname = BoolOption('openid', 'use_nickname_as_authname', False,
+            """ Whether the nickname as retrieved by SReg is used as username""")
+
+    trust_authname = BoolOption('openid', 'trust_authname', False,
+            """WARNING: Only enable this if you know what this mean!
+            This could make identity theft very easy if you do not control the OpenID provider!
+            Enabling this option makes the retrieved authname from the
+            OpenID provider authorative, i.e. it trusts the authname
+            to be the unique username of the user. Enabling this disables
+            the collision checking, so it may be possible to take over
+            an existing username or permission group.""")
+
     pape_method = Option('openid', 'pape_method', None,
             """Default PAPE method to request from OpenID provider.""")
 
@@ -180,6 +198,12 @@ class AuthOpenIdPlugin(Component):
     custom_provider_size = Option('openid', 'custom_provider_size', 'small',
             """ Custom OpenId provider image size (small or large).""")
 
+    groups_to_request = ListOption('openid', 'groups_to_request', '',
+            doc=""" Which 'team names' to request via the OpenIDTeams extension.
+            To use this option you must have python-openid-teams installed.
+            """)
+
+
 
     def _get_masked_address(self, address):
         if self.check_ip:
@@ -231,6 +255,12 @@ class AuthOpenIdPlugin(Component):
         if type(store) is not MemoryStore:
             store.createTables()
 
+    # IPermissionGroupProvider methods
+    def get_permission_groups(self, username):
+        ds = DetachedSession(self.env, username)
+        return ds.get('openid.teams', '').split(',')
+
+
     # IEnvironmentSetupParticipant methods
 
     def environment_created(self):
@@ -327,7 +357,7 @@ class AuthOpenIdPlugin(Component):
             'custom_provider_size': self.custom_provider_size,
             }, None
 
-    def _get_session(self, req):
+    def _get_oidsession(self, req):
         """Returns a session dict that can store any kind of object."""
         try:
             return cPickle.loads(str(req.session[self.openid_session_key]))
@@ -346,11 +376,11 @@ class AuthOpenIdPlugin(Component):
 
         return base_url
 
-    def _commit_session(self, session, req):
-        req.session[self.openid_session_key] = str(cPickle.dumps(session))
+    def _commit_oidsession(self, oidsession, req):
+        req.session[self.openid_session_key] = str(cPickle.dumps(oidsession))
 
     def _get_consumer(self, req, db):
-        s = self._get_session(req)
+        s = self._get_oidsession(req)
         if 'id' not in s:
             s['id'] = req.session.sid
         store = self._getStore(db)
@@ -385,7 +415,7 @@ class AuthOpenIdPlugin(Component):
         immediate = 'immediate' in req.args
 
         db = self.env.get_db_cnx()
-        oidconsumer, session = self._get_consumer(req, db)
+        oidconsumer, oidsession = self._get_consumer(req, db)
         try:
             self.env.log.debug('beginning OpenID authentication.')
             request = oidconsumer.begin(openid_url)
@@ -425,7 +455,7 @@ class AuthOpenIdPlugin(Component):
                     'custom_provider_size': self.custom_provider_size,
                    }, None
             else:
-                self._commit_session(session, req)
+                self._commit_oidsession(oidsession, req)
                 # Then, ask the library to begin the authorization.
                 # Here we find out the identity server that will verify the
                 # user's identity, and get a token that allows us to
@@ -451,6 +481,8 @@ class AuthOpenIdPlugin(Component):
                     sreg_req = sreg_fields
                 else:
                     sreg_opt = sreg_fields
+                if self.use_nickname_as_authname:
+                    sreg_req.append('nickname')
                 sreg_request = sreg.SRegRequest(optional=sreg_opt, required=sreg_req)
                 request.addExtension(sreg_request)
 
@@ -460,6 +492,16 @@ class AuthOpenIdPlugin(Component):
                     ax_request.add(attr_info)
                 request.addExtension(ax_request)
 
+                if self.groups_to_request:
+                    if not TeamsRequest:
+                        self.env.log.error(
+                            'The python-openid-teams package is not installed.'
+                            ' The groups_to_request configuration option will'
+                            ' be ignored.')
+                    else:
+                        request.addExtension(
+                            TeamsRequest(requested=self.groups_to_request))
+
                 trust_root = self._get_trust_root(req)
                 if self.absolute_trust_root:
                     trust_root += '/'
@@ -482,17 +524,50 @@ class AuthOpenIdPlugin(Component):
                         'form': form_html
                        }, None
 
+    def get_user(self, openid_identifier):
+        """ Look up username by OpenID identifier
+
+        In the case that multiple users match, the one who has most
+        recently logged in will be returned.
+
+        :returns: username or ``None``.
+        """
+        db = self.env.get_db_cnx()
+        cursor = db.cursor()
+        cursor.execute("SELECT session.sid"
+                  " FROM session"
+                  " INNER JOIN session_attribute AS attr"
+                  "                  USING(sid, authenticated)"
+                  " WHERE session.authenticated=%s"
+                  "       AND attr.name=%s AND attr.value=%s"
+                  " ORDER BY session.last_visit DESC",
+                  (1, self.openid_session_identity_url_key, openid_identifier))
+        rows = cursor.fetchall()
+        if len(rows) == 0:
+            return None
+        elif len(rows) > 1:
+            # Multiple users matched.  (We will return the one who most
+            # recently logged in.)
+
+            # FIXME: Probably should provide a config option which
+            # controls whether this is a error or not.
+            self.log.warning(
+                "Multiple users share the same openid identifier: %s",
+                ', '.join("'%s'" % user for (user,) in rows))
+        return rows[0][0]
+
     def _do_process(self, req):
         """Handle the redirect from the OpenID server.
         """
         db = self.env.get_db_cnx()
-        oidconsumer, session = self._get_consumer(req, db)
+        oidconsumer, oidsession = self._get_consumer(req, db)
 
         # Ask the library to check the response that the server sent
         # us.  Status is a code indicating the response type. info is
         # either None or a string containing more information about
         # the return type.
-        info = oidconsumer.complete(req.args,req.args['openid.return_to'])
+        current_url = req.abs_href(req.path_info)
+        info = oidconsumer.complete(req.args,current_url)
 
         css_class = 'error'
         if info.status == consumer.FAILURE and info.identity_url:
@@ -508,6 +583,8 @@ class AuthOpenIdPlugin(Component):
             # the verification.
             css_class = 'alert'
 
+            session_attr = {}           # attributes for new "user"
+
             # This is a successful verification attempt. If this
             # was a real application, we would do our login,
             # comment posting, etc. here.
@@ -529,10 +606,23 @@ class AuthOpenIdPlugin(Component):
                      or ax_info.get('email2')
                      or sreg_info.get('email'))
 
-            fullname = (' '.join(filter(None, map(ax_info.get,
-                                                  ('firstname', 'lastname'))))
-                        or sreg_info.get('fullname')
-                        or email.split('@',1)[0].replace('.', ' ').title())
+            fullname = (
+                ' '.join(filter(None, map(ax_info.get,
+                                          ('firstname', 'lastname'))))
+                or sreg_info.get('fullname')
+                or (email and email.split('@',1)[0].replace('.', ' ').title()))
+
+            nickname = sreg_info.get('nickname')
+
+            if self.groups_to_request and TeamsResponse:
+                teams_response = TeamsResponse.fromSuccessResponse(info)
+                if teams_response:
+                    # be careful not to make user a member of any trac groups
+                    # not named in groups_to_request
+                    teams = set(teams_response.teams
+                                ).intersection(self.groups_to_request)
+                    if teams:
+                        session_attr['openid.teams'] = ','.join(teams)
 
             if self.strip_protocol:
                 remote_user = remote_user[remote_user.find('://')+3:]
@@ -561,27 +651,33 @@ class AuthOpenIdPlugin(Component):
                     if item.match(remote_user):
                         allowed = False
                         self.env.log.debug("User black-listed.")
-            if allowed and email and self.re_email_white_list:
-                self.env.log.debug("Filtering email '%s' through email white-list." % email)
+            if allowed and self.re_email_white_list:
+                self.env.log.debug("Filtering email %r through email white-list." % email)
                 allowed = False
-                for item in self.re_email_white_list:
-                    if not allowed and item.match(email):
-                        allowed = True
-                        self.env.log.debug("User email white-listed.")
+                if email:
+                    for item in self.re_email_white_list:
+                        if not allowed and item.match(email):
+                            allowed = True
+                            self.env.log.debug("User email white-listed.")
 
             if allowed and self.check_list:
+                allowed = False
                 params = {self.check_list_key: remote_user}
                 if email:
                     params['email'] = email
                 url = self.check_list + '?' + urllib.urlencode(params)
                 self.env.log.debug('OpenID check list URL: %s' % url)
-                result = json.load(urllib.urlopen(url))
-                if not result[self.check_list_key]:
-                    allowed = False
-                elif self.check_list_username:
-                    new_user = result[self.check_list_username]
-                    if new_user:
-                        remote_user = new_user
+                try:
+                    result = json.load(urllib.urlopen(url))
+                    if result[self.check_list_key]:
+                        if self.check_list_username:
+                            cl_username = unicode(
+                                result[self.check_list_username])
+                            if not cl_username:
+                                raise ValueError("Bad value for username")
+                        allowed = True
+                except Exception, ex:
+                    self.env.log.error('OpenID check_list failed: %s' % ex)
 
             if allowed:
                 cookie = hex_entropy()
@@ -592,46 +688,81 @@ class AuthOpenIdPlugin(Component):
                 if cookie_lifetime > 0:
                     req.outcookie['trac_auth']['expires'] = cookie_lifetime
 
-                req.session[self.openid_session_identity_url_key] = info.identity_url
+                session_attr[self.openid_session_identity_url_key] = info.identity_url
                 if email:
-                    req.session['email'] = email
+                    session_attr['email'] = email
                 if fullname:
-                    req.session['name'] = fullname
+                    session_attr['name'] = fullname
 
-                self._commit_session(session, req)
+                self._commit_oidsession(oidsession, req)
 
-                if req.session.get('name'):
-                    authname = req.session['name']
-                    if self.combined_username:
-                        authname = '%s <%s>' % (authname, remote_user)
-
-                # Possibly lower-case the authname.
-                if self.lowercase_authname:
-                    authname = authname.lower()
-
-                # Make authname unique in case of collisions
-                #
-                # XXX: We ought to first look for an existing authenticated
-                # ssession with matching identity_url, and just use that
-                # for the authid.  (E.g. what if the user changes his
-                # fullname at the openid provider?)  However, trac does
-                # not seem to provide an API for searching sessions other
-                # than by sid/authname.
-                #
-                def authnames(base):
-                    yield base
-                    for attempt in itertools.count(2):
-                        yield "%s (%d)" % (base, attempt)
-
-                for authname in authnames(authname):
+                # First look for an existing authenticated session with
+                # matching identity_url.
+                self.env.log.debug('Checking URL: %s' % info.identity_url)
+                authname_for_identity_url = self.get_user(info.identity_url)
+                if authname_for_identity_url:
+                    authname = authname_for_identity_url
                     ds = DetachedSession(self.env, authname)
-                    if ds.last_visit == 0 and len(ds) == 0:
-                        # At least in 0.12.2, this mean no session exists.
-                        break
-                    ds_identity = ds.get(self.openid_session_identity_url_key)
-                    if ds_identity == info.identity_url:
-                        # No collision
-                        break
+                    # The user already exists, update team membership
+                    # XXX: Should also update name and/or email? (This would
+                    # be an API change.)
+                    for name in ['openid.teams']:
+                        if name in session_attr:
+                            ds[name] = session_attr[name]
+                        elif name in ds:
+                            del ds[name]
+                    ds.save()
+                else:
+                    # New identity URL -> create new authname/user.
+                    if self.check_list and self.check_list_username:
+                        authname = cl_username
+                    elif self.use_nickname_as_authname and nickname:
+                        authname = nickname
+                    elif session_attr.get('name'):
+                        authname = session_attr['name']
+                        if self.combined_username:
+                            authname = '%s <%s>' % (authname, remote_user)
+                    else:
+                        authname = remote_user
+
+                    # Possibly lower-case the authname.
+                    if self.lowercase_authname:
+                        authname = authname.lower()
+
+                    if self.trust_authname:
+                        ds = DetachedSession(self.env, authname)
+                    else:
+                        # Make authname unique in case of collisions
+                        def authnames(base):
+                            yield base
+                            for attempt in itertools.count(2):
+                                yield "%s (%d)" % (base, attempt)
+
+                        users_and_groups_with_permissions = set(
+                            user
+                            for user, perm
+                            in PermissionSystem(self.env).get_all_permissions())
+
+                        for authname in authnames(authname):
+                            ds = DetachedSession(self.env, authname)
+                            # At least in 0.12.2, this means no session exists.
+                            no_session_exists = ds.last_visit == 0 and len(ds) == 0
+                            no_permissions_defined = authname not in users_and_groups_with_permissions
+                            if (no_session_exists and no_permissions_defined):
+                                # name is free :-)
+                                break
+                        # Set attributes for new user on the
+                        # current anonymous session.  It will be promoted to
+                        # the new authenticated session on the next request
+                        # (by Session.__init__).
+                        #
+                        # NB: avoid dict.update here to ensure that
+                        # DetachedSession.__getitem__ gets a chance to
+                        # normalize values
+                        for name, value in session_attr.items():
+                            req.session[name] = value
+                        self.env.log.info("Created new user '%s' for "
+                            "OpenID identifier %s", authname, info.identity_url)
 
                 req.authname = authname
 
@@ -663,7 +794,7 @@ class AuthOpenIdPlugin(Component):
             # information in a log.
             message = 'Verification failed.'
 
-        self._commit_session(session, req)
+        self._commit_oidsession(oidsession, req)
 
         add_stylesheet(req, 'authopenid/css/openid.css')
         add_script(req, 'authopenid/js/openid-jquery.js')
